# Generated by Django 2.2.8 on 2020-01-11 18:53

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="Context",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("object_id", models.PositiveIntegerField()),
                (
                    "content_type",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="contenttypes.ContentType",
                    ),
                ),
            ],
            options={"unique_together": {("content_type", "object_id")}},
        ),
        migrations.CreateModel(
            name="Role",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("definition_json", models.TextField()),
                (
                    "rbac_context",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="scoped_rbac.Context",
                    ),
                ),
            ],
            options={"abstract": False},
        ),
        migrations.CreateModel(
            name="RoleAssignment",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "rbac_context",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="scoped_rbac.Context",
                    ),
                ),
                (
                    "role",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to="scoped_rbac.Role",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={"abstract": False},
        ),
    ]



# Generated by Django 2.2.17 on 2021-01-04 16:14

from django.db import migrations, models
from uuid import uuid4
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('scoped_rbac', '0004_auto_20200223_1921'),
    ]

    operations = [
        migrations.AddField(
            model_name='role',
            name='etag',
            field=models.CharField(default=uuid4().hex, max_length=40),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='role',
            name='modified_at',
            field=models.DateTimeField(default=django.utils.timezone.now),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='roleassignment',
            name='etag',
            field=models.CharField(default=uuid4().hex, max_length=40),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='roleassignment',
            name='modified_at',
            field=models.DateTimeField(default=django.utils.timezone.now),
            preserve_default=False,
        ),
    ]



"""
RBAC Policies, with stubbed-out support for conditional expressions.
"""

from collections import namedtuple
import json


Permission = namedtuple("Permission", "action, resource_type")


class Policy:
    def should_allow(self, *args, **kwargs):
        raise NotImplementedError()

    def sum_with(self, other_policy):
        raise NotImplementedError()

    def __repr__(self):
        raise NotImplementedError()

    def to_json(self):
        raise NotImplementedError()


class PolicyBoolean(Policy):
    ...


class PolicyTrue(PolicyBoolean):
    def should_allow(self, *args, **kwargs):
        return True

    def sum_with(self, other_policy):
        return self

    def __repr__(self):
        return "PolicyTrue"

    def to_json(self):
        return True


class PolicyFalse(PolicyBoolean):
    def should_allow(self, *args, **kwargs):
        return False

    def sum_with(self, other_policy):
        return other_policy

    def __repr__(self):
        return "PolicyFalse"

    def to_json(self):
        return False


POLICY_TRUE = PolicyTrue()
POLICY_FALSE = PolicyFalse()


class Expression(Policy):
    """
    Expression policies are initialized with a `dict` detailing the parameters
    to use in evaluating the expression to determine whether the policy
    conditions are met.
    """

    def __init__(self, expression):
        self.expression = expression

    def evaluate(self, permission, context_id, resource=None):
        raise NotImplementedError()

    def should_allow(self, permission, context_id, resource=None):
        return self.evaluate(self, permission, context_id, resource)

    def sum_with(self, other_policy):
        if isinstance(other_policy, PolicyBoolean):
            return other_policy.sum_with(self)
        if isinstance(other_policy, Expression):
            return ExpressionList(self, other_policy)
        if isinstance(other_policy, PolicySet):
            return CompoundPolicy(expression=self, policy_set=other_policy)
        if isinstance(other_policy, PolicyDict):
            return CompoundPolicy(expression=self, policy_dict=other_policy)
        if isinstance(other_policy, CompoundPolicy):
            return other_policy.add_expression(self)

    def __repr__(self):
        return f"Expression {repr(self.expression)}"

    def to_json(self):
        return self.expression

    @classmethod
    def from_json(cls, json_policy):
        return Expression(json_policy)


class ExpressionList(Policy):
    def __init__(self, *args):
        self.expressions = args

    def should_allow(self, *args, **kwargs):
        for expression in self.expressions:
            if expression.should_allow(*args, **kwargs):
                return True
        return False

    def sum_with(self, other_policy):
        if isinstance(other_policy, PolicyBoolean):
            return other_policy.sum_with(self)
        if isinstance(other_policy, Expression):
            return self.add(other_policy)
        if isinstance(other_policy, PolicySet):
            return CompoundPolicy(expression=self, policy_set=other_policy)
        if isinstance(other_policy, PolicyDict):
            return CompoundPolicy(expression=self, policy_dict=other_policy)
        if isinstance(other_policy, CompoundExpression):
            return other_policy.add_expression(self)

    def add(self, expression):
        return ExpressionList(expression, *self.expressions)

    def __repr__(self):
        return f"ExpressionList [ {', '.join([repr(expr) for expr in self.expressions])}  ]"

    def to_json(self):
        return [expr.to_json() for expr in self.expressions]

    @classmethod
    def from_json(cls, json_policy):
        return ExpressionList(*[Expression.from_json(item) for item in json_policy])


class PolicySet(Policy):
    def __init__(self, *args):
        self.allowed = set(args)

    def should_allow(self, *args, **kwargs):
        if len(args) is 0:
            return False
        key = args[0]
        return key in self.allowed

    def sum_with(self, other_policy):
        if (
            isinstance(other_policy, PolicyBoolean)
            or isinstance(other_policy, Expression)
            or isinstance(other_policy, ExpressionList)
        ):
            return other_policy.sum_with(self)
        if isinstance(other_policy, PolicySet):
            return PolicySet(*self.allowed.union(other_policy.allowed))
        if isinstance(other_policy, PolicyDict):
            return other_policy.add_all([(key, POLICY_TRUE) for key in self.allowed])
        if isinstance(other_policy, CompoundPolicy):
            return other_policy.add_policy_set(self)
        raise NotImplementedError(f"Unsupported type {type(other_policy)}")

    def __repr__(self):
        return f"PolicySet {repr(self.allowed)}"

    def to_json(self):
        """Convert to JSON. Returns a sorted list so that the representation of the set
        may be stable for testing and other purposes.
        """
        return sorted(list(self.allowed))

    @classmethod
    def from_json(cls, json_policy):
        if isinstance(json_policy, str):
            return PolicySet(json_policy)
        return PolicySet(*json_policy)


class PolicyDict(Policy):
    def __init__(self, policy_dict):
        self.policies = policy_dict

    def should_allow(self, *args, **kwargs):
        if len(args) is 0:
            return False
        key = args[0]
        policy = self.policies.get(key, POLICY_FALSE)
        return policy.should_allow(*args[1:], *kwargs)

    def sum_with(self, other_policy):
        if (
            isinstance(other_policy, PolicyBoolean)
            or isinstance(other_policy, Expression)
            or isinstance(other_policy, ExpressionList)
        ):
            return other_policy.sum_with(self)
        if isinstance(other_policy, PolicySet):
            return other_policy.sum_with(self)
        if isinstance(other_policy, PolicyDict):
            return self.recursive_sum_with(other_policy)
        if isinstance(other_policy, CompoundPolicy):
            return other_policy.add_policy_set(self)
        raise NotImplementedError(f"Unsupported type {type(other_policy)}")

    def add_all(self, key_policy_pairs):
        policies = dict(self.policies)
        for key, policy in key_policy_pairs:
            current = policies.get(key, POLICY_FALSE)
            policies[key] = policy.sum_with(current)
        return PolicyDict(policies)

    def recursive_sum_with(self, other_policy):
        policies = dict(self.policies)
        for k, v in other_policy.policies.items():
            current_policy = policies.get(k, POLICY_FALSE)
            policies[k] = current_policy.sum_with(v)
        return PolicyDict(policies)

    def __repr__(self):
        return f"PolicyDict {repr(self.policies)}"

    def to_json(self):
        return {k: v.to_json() for k, v in self.policies.items()}

    @classmethod
    def from_json(cls, json_policy):
        policies = {key: policy_from_json(value) for key, value in json_policy.items()}
        return PolicyDict(policies)

    def keys(self):
        return self.policies.keys()


class CompoundPolicy(Policy):
    def __init__(self, policy_dict=None, policy_set=None, expressions=None):
        self.policy_dict = policy_dict or PolicyDict({})
        if policy_set:
            self.policy_dict = self.policy_dict.sum_with(policy_set)
        self.expressions = expressions or ExpressionList()

    def should_allow(self, *args, **kwargs):
        return self.policy_dict.should_allow(
            *args, **kwargs
        ) or self.expressions.should_allow(*args, **kwargs)

    def sum_with(self, other_policy):
        return other_policy.sum_with(self)

    def add_expression(self, expression):
        return CompoundPolicy(
            policy_dict=self.policy_dict,
            expressions=self.expressions.sum_with(expression),
        )

    def add_policy_set(self, policy_set):
        return CompoundPolicy(
            policy_dict=self.policy_dict,
            policy_set=policy_set,
            expressions=self.expressions,
        )

    def add_policy_dict(self, policy_dict):
        return CompoundPolicy(
            policy_dict=self.policy_dict.sum_with(policy_dict),
            expressions=self.expressions,
        )

    def __repr__(self):
        return (
            f"CompoundPolicy {{ expressions: {repr(self.expressions)}, "
            f"policy_dict: {repr(self.policy_dict)} }}"
        )

    def to_json(self):
        ret = dict()
        if self.expressions is not None:
            ret["expressions"] = expressions.to_json()
        if self.policy_dict is not None and self.policy_dict.keys():
            ret["policy_dict"] = policy_dict.to_json()
        return ret

    @classmethod
    def from_json(cls, json_policy):
        expressions = None
        policy_dict = None
        if "expressions" in json_policy:
            expressions = ExpressionList.from_json(json_policy["expressions"])
        if "policy_dict" in json:
            policy_dict = PolicyDict.from_json(json_policy["policy_dict"])
        return CompoundPolicy(policy_dict=policy_dict, expressions=expressions)


def policy_from_json(json_policy):
    if json_policy is True:
        return POLICY_TRUE
    if json_policy is False:
        return POLICY_FALSE
    if isinstance(json_policy, list) or isinstance(json_policy, str):
        return PolicySet.from_json(json_policy)
    if isinstance(json_policy, dict):
        if "expressions" in json_policy or "policy_dict" in json_policy:
            return CompoundPolicy.from_json(json_policy)
        return PolicyDict.from_json(json_policy)


class RootPolicy:
    def __init__(self):
        self.policy = POLICY_FALSE

    def should_allow(self, permission, context_id, resource=None):
        return self.policy.should_allow(context_id, *permission, resource=resource)

    def add_json_policy_for_context(self, json_policy, context):
        policy = policy_from_json(json_policy)
        self.add_policy(PolicyDict({context: policy}))
        return self

    def add_policy_for_context(self, policy, context):
        self.add_policy(PolicyDict({context: policy}))
        return self

    def add_policy(self, policy):
        self.policy = self.policy.sum_with(policy)
        return self

    def __repr__(self):
        return f"RootPolicy {repr(self.policy)}"

    def to_json(self):
        return self.policy.to_json()

    def get_contexts_for(self, permission):
        # FIXME This isn't done, yet
        if isinstance(self.policy, PolicyDict):
            return self.policy.keys()
        if isinstance(self.policy, CompoundPolicy) and self.policy.expressions is None:
            return self.policy.policy_dict.keys()
        return None



from django.contrib import admin

from .models import Role, RoleAssignment


@admin.register(Role)
class RoleAdmin(admin.ModelAdmin):
    ...


@admin.register(RoleAssignment)
class RoleAssignmentAdmin(admin.ModelAdmin):
    ...



from rest_framework.authentication import BaseAuthentication
from .conf import policy_for_unauthenticated
from .permissions import http_action_iri_for
from .rbac_contexts import SOME_CONTEXT


class ConditionalAnonymousAuthentication(BaseAuthentication):
    def authenticate(self, request):
        if "HTTP_AUTHORIZATION" in request.META:
            return None
        policy = policy_for_unauthenticated()
        if policy.should_allow(
                policy.Permission(
                    http_action_iri_for(request), view.resource_type_iri_for(request)
                ),
                SOME_CONTEXT,
                None):
            return AnonymousUser(), None
        return None



# Generated by Django 2.2.8 on 2020-02-19 14:05

from django.db import migrations
import scoped_rbac.fields


class Migration(migrations.Migration):

    dependencies = [("scoped_rbac", "0001_initial")]

    operations = [
        migrations.RenameField(
            model_name="role", old_name="rbac_context", new_name="context"
        ),
        migrations.RenameField(
            model_name="roleassignment", old_name="rbac_context", new_name="context"
        ),
        migrations.RemoveField(model_name="role", name="definition_json"),
        migrations.AddField(
            model_name="role",
            name="definition",
            field=scoped_rbac.fields.JSONField(default={}),
            preserve_default=False,
        ),
    ]



from collections import namedtuple
from dataclasses import dataclass
from typing import List


Action = namedtuple("Action", "iri display_name, description")


@dataclass
class ResourceType:
    iri: str
    display_name: str
    description: str

    @property
    def list_iri(self) -> str:
        return f"{self.iri}/list"


class RbacRegistry:
    ACTIONS = list()
    CACHED_RESOURCE_TYPES = list()

    _processed_model_classes = False
    _ACCESS_CONTROLLED_MODEL_CLASSES = list()

    @classmethod
    def known_resource_types(cls):
        if not cls._processed_model_classes:
            cls._processed_model_classes = True
            for model_cls in cls._ACCESS_CONTROLLED_MODEL_CLASSES:
                if not model_cls._meta.abstract:
                    if not hasattr(model_cls, "resource_type"):
                        raise Exception(
                            "Subclasses of AccessControlled **MUST** have a "
                            f"`resource_type: ResourceType` property. {model_cls}"
                        )
                    cls.CACHED_RESOURCE_TYPES.append(model_cls.resource_type)
        return cls.CACHED_RESOURCE_TYPES


def register_access_controlled_model(cls):
    RbacRegistry._ACCESS_CONTROLLED_MODEL_CLASSES.append(cls)


def register_resource_types(*args: List[ResourceType]):
    RbacRegistry.CACHED_RESOURCE_TYPES.extend(args)


def register_actions(*actions: List[Action]):
    RbacRegistry.ACTIONS.extend(*actions)



# Generated by Django 2.2.8 on 2020-02-23 19:21

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [("scoped_rbac", "0003_auto_20200223_1846")]

    operations = [
        migrations.AddField(
            model_name="role",
            name="rbac_context",
            field=models.CharField(default="", max_length=2048),
        ),
        migrations.AddField(
            model_name="roleassignment",
            name="rbac_context",
            field=models.CharField(default="", max_length=2048),
        ),
    ]



from django.contrib.auth import get_user_model
from django.utils.timezone import now
from rest_framework import serializers
from uuid import uuid4
from .models import Role, RoleAssignment


class RoleSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = Role
        # fields = ("definition", )
        fields = "__all__"

    etag = serializers.HiddenField(default=lambda: uuid4().hex)
    modified_at = serializers.HiddenField(default=now)


class RoleAssignmentSerializer(serializers.HyperlinkedModelSerializer):
    class Meta:
        model = RoleAssignment
        # fields = ("user", "role", "resource_type", )
        fields = "__all__"

    etag = serializers.HiddenField(default=lambda: uuid4().hex)
    modified_at = serializers.HiddenField(default=now)


class UserSerializer(serializers.HyperlinkedModelSerializer):
    """User Serialization"""

    class Meta:
        model = get_user_model()
        fields = ("url", "username", "email", "first_name", "last_name", "id", )



from django.conf import settings
from django.db import models
from .policy import policy_from_json
from .fields import JSONField
from .registry import ResourceType, register_access_controlled_model


class IdentifiedByIRI(object):
    """
    A model mixin that has an associated RDF IRI indicating its RDF type.

    Subclasses **MUST** define a `resource_type: ResourceType` property.
    """

    ...


class AccessControlledModel(IdentifiedByIRI, models.Model):
    """
    Model classes that will be access controlled in a `rest_framework` view **MUST**
    subclass this class.

    Subclasses **MUST** define a `resource_type: ResourceType` property.
    """

    rbac_context = models.CharField(max_length=2048, default="")

    class Meta:
        abstract = True

    @classmethod
    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        register_access_controlled_model(cls)


class Role(AccessControlledModel):
    definition = JSONField(null=False)
    modified_at = models.DateTimeField()
    etag = models.CharField(max_length=40)

    # Required by AccessControlled
    resource_type = ResourceType(
        "rbac.Role", "Role", "A Role definition as a JSON resource."
    )

    @property
    def as_policy(self):
        return policy_from_json(self.definition)


class RoleAssignment(AccessControlledModel):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    role = models.ForeignKey(Role, on_delete=models.CASCADE)
    resource_type = ResourceType(
        "rbac.RoleAssignment",
        "RoleAssignment",
        "The assignment of an rbac.Role to a User.",
    )
    modified_at = models.DateTimeField()
    etag = models.CharField(max_length=40)


# TODO Figure out how to support custom User models
UserResourceType = ResourceType("rbac.User", "User", "A resource representing a User.")



from rest_framework.filters import BaseFilterBackend
from .permissions import http_action_iri_for, policy_for
from .policy import Permission


class RbacFilter(BaseFilterBackend):
    """
    Filter that only allows users to see the objects they're authorized to access for
    the request.
    """

    def filter_queryset(self, request, queryset, view):
        if request.method == "GET":
            return self.filter_queryset_for_get(request, queryset, view)
        else:
            return queryset

    def filter_queryset_for_get(self, request, queryset, view):
        policy = policy_for(request)
        contexts = policy.get_contexts_for(
            Permission(
                http_action_iri_for(request), view.resource_type_iri_for(request)
            )
        )
        if contexts is None:
            return queryset
        if not contexts:
            return queryset.none()
        return queryset.filter(rbac_context__in=contexts)



# Generated by Django 2.2.8 on 2020-02-23 18:46

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [("scoped_rbac", "0002_auto_20200219_1405")]

    operations = [
        migrations.RemoveField(model_name="role", name="context"),
        migrations.RemoveField(model_name="roleassignment", name="context"),
        migrations.DeleteModel(name="Context"),
    ]



from django.conf import settings
from rest_framework.settings import import_from_string


DEFAULTS = {
    "POLICY_FOR_AUTHENTICATED": "scoped_rbac.policy.POLICY_FALSE",
    "POLICY_FOR_STAFF": "scoped_rbac.policy.POLICY_FALSE",
    "POLICY_FOR_UNAUTHENTICATED": "scoped_rbac.policy.POLICY_FALSE",
}

IMPORT_STRINGS = ["POLICY_FOR_UNAUTHENTICATED", "POLICY_FOR_STAFF", ]


def scoped_rbac_settings():
    return getattr(settings, "SCOPED_RBAC", dict())


def import_policy_or_func(setting):
    default = DEFAULTS[setting]
    policy_or_func = scoped_rbac_settings().get(setting, default)
    if isinstance(policy_or_func, str):
        policy_or_func = import_from_string(policy_or_func, setting)
        if callable(policy_or_func):
            policy_or_func = policy_or_func()
    return policy_or_func


def policy_for_unauthenticated():
    """The default policy for unauthenticated users may be defined via settings. The
    value in settings may be either a Policy instance or a string fully qualified name
    of a function that returns a Policy instance. For example:

    .. code-block:: python

       SCOPED_RBAC = {
         "POLICY_FOR_UNAUTHENTICATED": "my_app.default_policy_for_unauthenticated",
         ...
       }
    """
    return import_policy_or_func("POLICY_FOR_UNAUTHENTICATED")


def policy_for_authenticated():
    """The base policy for authenticated users may be defined via settings. The value in
    seetgings may be either a Policy instance or a string fully qualified name of a
    function that returns a Policy instance. For example:

    .. codeblock:: python

       SCOPED_RBAC = {
         "POLICY_FOR_AUTHENTICATED": "my_app.base_policy_for_authenticated",
         ...
       }
    """
    return import_policy_or_func("POLICY_FOR_AUTHENTICATED")


def policy_for_staff():
    """The default policy for staff users may be defined via settings. The value in
    settings may be ether a Policy instance or a string fully qualified name of a
    function that returns a Policy instance. For example:

    .. code-block:: python

       SCOPED_RBAC = {
         "POLICY_FOR_STAFF": "my_app.default_policy_for_staff"
       }
    """
    return import_policy_or_func("POLICY_FOR_STAFF")



from rest_framework import permissions
from .models import RoleAssignment
from . import policy
from .rbac_contexts import DEFAULT_CONTEXT, SOME_CONTEXT
from .conf import policy_for_authenticated, policy_for_staff, policy_for_unauthenticated


NOT_ALLOWED = policy.RootPolicy().add_policy(policy.POLICY_FALSE)
ALLOWED = policy.RootPolicy().add_policy(policy.POLICY_TRUE)


def sum_policies(total_policy, additional_policy):
    """Add the `additional_policy` to the `total_policy` and also cache the permissions
    granted by `additional_policy` as also granted in `SOME_CONTEXT`.
    """
    total_policy.add_policy(additional_policy)
    if isinstance(additional_policy, policy.PolicyDict):
        policy_dict = additional_policy
    elif isinstance(additional_policy, policy.CompoundPolicy):
        policy_dict = additional_policy.policy_dict
    else:
        policy_dict = policy.PolicyDict({})
    for context, p in policy_dict.policies.items():
        total_policy.add_policy_for_context(p, SOME_CONTEXT)


def policy_for(request):
    # TODO figure out caching for this
    if request.user.is_superuser:
        return ALLOWED
    total_policy = policy.RootPolicy()
    if request.user is None or request.user.is_anonymous:
        sum_policies(total_policy, policy_for_unauthenticated())
    else:
        sum_policies(total_policy, policy_for_authenticated())
        if request.user.is_staff:
            sum_policies(total_policy, policy_for_staff())
        role_assignments = RoleAssignment.objects.filter(
            user=request.user
        ).prefetch_related("role")
        policy_by_role = dict()
        for role_assignment in role_assignments.all():
            role = role_assignment.role
            if role not in policy_by_role:
                policy_by_role[role] = role.as_policy
            policy_for_role = policy_by_role[role]
            total_policy.add_policy_for_context(
                policy_for_role, role_assignment.rbac_context
            )
            total_policy.add_policy_for_context(policy_for_role, SOME_CONTEXT)
    return total_policy


def http_action_iri_for(request):
    method = request.method
    if method == "HEAD":
        method = "GET"
    return f"http.{method}"


class IsAuthorized(permissions.BasePermission):
    """
    Custom permission handling using the `rbac` model.
    """

    def has_object_permission(self, request, view, obj):
        """
        Requires that the object is `AccessControlled`.
        """
        if not hasattr(view, "resource_type_iri_for"):
            return True
        effective_policy = policy_for(request)
        permission = policy.Permission(
                http_action_iri_for(request), obj.resource_type.iri),
        permitted = effective_policy.should_allow(
            policy.Permission(http_action_iri_for(request), obj.resource_type.iri),
            obj.rbac_context,
            obj,
        )
        return permitted

    def has_permission(self, request, view):
        """
        Requires the view to be an `AccessControlledAPIView`.
        """
        if not hasattr(view, "resource_type_iri_for"):
            return True
        effective_policy = policy_for(request)
        if request.method in ("PUT", "POST"):
            resource = request.data if request.method in ("PUT", "POST") else None
            rbac_context = resource.get("rbac_context", DEFAULT_CONTEXT)
        else:
            resource = None
            rbac_context = SOME_CONTEXT
        permission = policy.Permission(
            http_action_iri_for(request), view.resource_type_iri_for(request)
        )
        permitted = effective_policy.should_allow(
            permission,
            rbac_context,
            resource,
        )
        return permitted



from django.urls import path
from rest_framework.routers import DefaultRouter
from . import rest

router = DefaultRouter()
router.register(r"role-assignments", rest.RoleAssignmentViewSet)
router.register(r"roles", rest.RoleViewSet)
router.register(r"users", rest.UserViewSet)

urlpatterns = router.urls + [
    path(
        "user-rbac-policy/", rest.UserRbacPolicyView.as_view(), name="user-rbac-policy"
    )
]



from django.contrib.auth import get_user_model
from django.shortcuts import get_object_or_404

# from django.urls import reverse
from rest_framework import status
from rest_framework.pagination import PageNumberPagination
from rest_framework.permissions import IsAuthenticated
from rest_framework.response import Response
from rest_framework.reverse import reverse
from rest_framework.views import APIView
from rest_framework.viewsets import ModelViewSet
from scrudful.rest_framework import scrudful_viewset
from .filters import RbacFilter
from .models import Role, RoleAssignment, UserResourceType
from .permissions import policy_for
from .serializers import RoleSerializer, RoleAssignmentSerializer, UserSerializer


class DefaultPageNumberPagination(PageNumberPagination):
    page_size = 100


class AccessControlledAPIView:
    """
    This `APIView` interface is required for non-object based views in combination with
    the `IsAuthorized` permission class.
    """

    filter_backends = [RbacFilter]

    def resource_type_iri_for(self, request):
        """
        Return the resource_type_iri for the resource type indicated by this request.
        """

        if hasattr(self, "action") and self.action == "list":
            return self.list_type_iri
        else:
            return self.resource_type_iri

    @property
    def resource_type_iri(self):
        """
        Subclasses **MUST** override this method.
        Return the resource_type_iri for the base resource supported by this view.
        """

        raise NotImplementedError()

    @property
    def list_type_iri(self):
        """
        Return the list resource_type_iri for this view.
        """

        return f"{self.resource_type_iri}/list"


class AccessControlledModelViewSet(AccessControlledAPIView, ModelViewSet):
    def get_success_headers(self, instance, request=None):
        try:
            return {
                "Location": reverse(
                    self.basename + "-detail", args=[instance.id], request=request
                )
            }
        except (TypeError, KeyError):
            return {}

    def list(self, request, *args, **kwargs):
        # TODO restrict listing to authorized contexts
        return super().list(request, *args, **kwargs)

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        self.perform_create(serializer)
        headers = self.get_success_headers(serializer.instance, request=request)
        # headers["etag"] = serializer.etag()
        # headers["last-modified"] = serializer.last_modified()
        # if serializer.link_header_content():
        # headers["link"] = serializer.link_header_content()
        return Response(
            serializer.data, status=status.HTTP_201_CREATED, headers=headers
        )

    def retrieve(self, request, *args, **kwargs):
        # TODO authorize
        return super().retrieve(request, *args, **kwargs)

    def update(self, request, *args, **kwargs):
        # TODO authorize
        return super().update(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        # TODO authorize
        return super().delete(request, *args, **kwargs)


@scrudful_viewset
class RoleViewSet(AccessControlledModelViewSet):
    """
    TODO restrict queryset to authorized contexts for user
    """

    class Meta:
        def etag_func(view_instance, request, pk: str):
            instance = get_object_or_404(Role, pk=int(pk))
            return instance.etag

        def last_modified_func(view_instance, request, pk: str):
            instance = get_object_or_404(Role, pk=int(pk))
            return instance.modified_at

    queryset = Role.objects.all()
    serializer_class = RoleSerializer
    pagination_class = DefaultPageNumberPagination

    @property
    def resource_type_iri(self):
        return Role.resource_type.iri


class RoleAssignmentViewSet(AccessControlledModelViewSet):
    """
    TODO restrict queryset to authorized contexts for user
    """

    class Meta:
        def etag_func(view_instance, request, pk: str):
            instance = get_object_or_404(RoleAssignment, pk=int(pk))
            return instance.etag

        def last_modified_func(view_instance, request, pk: str):
            instance = get_object_or_404(RoleAssignment, pk=int(pk))
            return instance.modified_at

    queryset = RoleAssignment.objects.all()
    serializer_class = RoleAssignmentSerializer
    pagination_class = DefaultPageNumberPagination

    @property
    def resource_type_iri(self):
        return RoleAssignment.resource_type.iri


class UserViewSet(AccessControlledModelViewSet):
    """
    API which allows users to be viewed or edited.
    TODO restrict queryset to authorized contexts for user
    """

    queryset = get_user_model().objects.all()
    serializer_class = UserSerializer
    pagination_class = DefaultPageNumberPagination

    @property
    def resource_type_iri(self):
        return UserResourceType.iri


class UserRbacPolicyView(APIView):
    permission_classes = [IsAuthenticated]

    @property
    def resource_type_iri(self):
        return "rbac.Policy"

    def get(self, request):
        user_policy = policy_for(request)
        return Response(user_policy.to_json())



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['scoped_rbac', 'scoped_rbac.migrations']

package_data = \
{'': ['*']}

install_requires = \
['django-rest-framework-condition>=0.1.0,<0.2.0',
 'django>=2.2.8,<3.0.0',
 'djangorestframework>=3.11.0,<4.0.0',
 'drf-extensions>=0.6.0,<0.7.0',
 'psycopg2>=2.8.6,<3.0.0',
 'scrudful>=0.1.0,<0.2.0']

setup_kwargs = {
    'name': 'django-scoped-rbac',
    'version': '0.3.1',
    'description': 'A rich and flexible Django application for role-based access control within distinct access control scopes supporting Django Rest Framework.',
    'long_description': None,
    'author': 'David Charboneau',
    'author_email': 'david@adadabase.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/openteams/django-scoped-rbac',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
